"""
Module for working with tog database
"""

import json
import os
import sqlite3
import time
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Union


import psycopg2
import pytz

from skit_labels.utils import to_datetime
from skit_labels import constants as const
from skit_labels.types import (
    AudioSegmentTask,
    CallTranscriptionTask,
    ConversationTask,
    DataGenerationTask,
    DictTask,
    SimulatedCallTask,
    Task,
)


def build_task(
    d: Dict, task_type: str, data_id: Optional[str] = None, tz=pytz.UTC
) -> Task:
    """
    Create a task from given data dictionary.
    """

    if task_type == "conversation":
        task = ConversationTask.from_dict(d)

        # Since the reftime from db is in UTC, we convert it to our timezone. This
        # is needed as saying 12 pm means different things in different timezones
        # and can't be translated without doing something stupid.
        task.reftime = to_datetime(task.reftime)
        task.reftime = task.reftime.astimezone(tz).isoformat()
    elif task_type == "simulated_call":
        task = SimulatedCallTask.from_dict(d)
    elif task_type == "audio_segment":
        task = AudioSegmentTask.from_dict(d)
    elif task_type == "dict":
        task = DictTask.from_dict(d, data_id)
    elif task_type == "call_transcription":
        task = CallTranscriptionTask.from_dict(d, data_id)
    elif task_type == "data_generation":
        task = DataGenerationTask.from_dict(d)
    else:
        raise TypeError(f"Invalid task type {task_type} provided.")
    return task


class SqliteDatabase:
    """
    Class mapping to a local sqlite database file which can keep only one job.
    """

    def __init__(self, filepath: str):
        self.filepath = filepath
        self._initialize()

    def _initialize(self):
        self.conn = sqlite3.connect(self.filepath)
        c = self.conn.cursor()
        c.execute(
            """CREATE TABLE IF NOT EXISTS data (
            data_id INTEGER NOT NULL,
            data TEXT NOT NULL,
            tag TEXT NOT NULL,
            is_gold BOOLEAN NOT NULL,
            tagged_time TEXT,
            job_id TEXT NOT NULL
        )"""
        )
        self.conn.commit()

    def insert_rows(self, rows: List):
        """
        Write rows in database. Each item of row is a tuple of following elements:
        - data_id : int
        - data: Dict
        - tag: Dict
        - is_gold: bool
        - tagged_time: Optional[str]
        - job_id: str
        """

        c = self.conn.cursor()
        c.executemany(
            "INSERT INTO data (data_id, data, tag, is_gold, tagged_time, job_id) VALUES (?, ?, ?, ?, ?, ?)",
            [(i, json.dumps(d), json.dumps(t), g, tt, ji) for i, d, t, g, tt, ji in rows],
        )
        self.conn.commit()


class Database:
    """
    Class holding connection with backend database.
    """

    def __init__(
        self,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
    ):
        self._initialize(db=db, user=user, password=password, host=host, port=port)

    def _initialize(
        self,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
    ):
        db = db or os.getenv(const.TOGDB_DB, "tog")
        user = user or os.getenv(const.TOGDB_USER)
        password = password or os.getenv(const.TOGDB_PASSWORD)
        host = host or os.getenv(const.TOGDB_HOST, "localhost")
        port = port or os.getenv(const.TOGDB_PORT, "5432")

        if password is None:
            raise ValueError(
                "Credentials for Tog database not set. Check for missing environment variables."
            )

        self.conn = psycopg2.connect(
            host=host, database=db, user=user, password=password, port=port
        )

    def list_jobs(self) -> List[Dict]:
        with self.conn.cursor() as cur:
            # NOTE: We are not picking out task_type field since that
            #       collides with our task type names. Ideally we need to
            #       settle on same nomenclature.
            cur.execute(
                "SELECT id, name, description, config, language FROM jobs_job WHERE is_active"
            )

            jobs: List[Dict] = []
            for row in cur.fetchall():
                jobs.append(
                    {
                        "id": row[0],
                        "name": row[1],
                        "description": row[2],
                        "config": row[3],
                        "language": row[4],
                    }
                )

            return jobs


class AbstractJob(ABC):
    """
    Abstract class representing connection to a tagging job.
    """

    @abstractmethod
    def total(self, untagged=False) -> int:
        ...

    @abstractmethod
    def get_by_data_id(self, id: int, cache=True):
        ...

    @abstractmethod
    def get(self, untagged=False, itersize=1000, only_gold=False):
        ...


class Job(AbstractJob):
    """
    A Tog job which specifies a kind of tagging data set and problem.
    """

    def __init__(
        self,
        id: int,
        task_type="conversation",
        database: Optional[Database] = None,
        tz=pytz.UTC,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[str, int]] = None,
    ):
        self.id = id
        # TODO: Check task validity right here
        self.task_type = task_type
        self.db_name = db
        self.db = database or Database(
            db=db, user=user, password=password, host=host, port=port
        )
        self._fetch_details()

        # Cache for keeping rows of job indexed by data_ids
        self.cache = {}
        self.tz = tz
        self.start_date = start_date
        self.end_date = end_date
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def __repr__(self):
        f"Job {self.id}: {self.name} [language: {self.lang}]\n{self.description}"

    def _fetch_details(self):
        """
        Save metadata about the current job id
        """

        with self.db.conn.cursor() as cur:
            # NOTE: We are not picking out task_type field since that
            #       collides with our task type names. Ideally we need to
            #       settle on same nomenclature.
            cur.execute(
                f"SELECT name, description, config, language FROM jobs_job WHERE id = {self.id}"
            )
            try:
                self.name, self.description, self.config, self.lang = cur.fetchone()
            except TypeError:
                raise ValueError("Invalid job id")

    def total(self, untagged=False, start_date=None, end_date=None):
        """
        Return total number of items for this job. If `untagged` is True, consider
        untagged items in counting too.
        """
        start_date = start_date or self.start_date
        end_date = end_date or self.end_date

        with self.db.conn.cursor() as cur:
            cur.execute(
                f"""
                SELECT count(*) FROM jobs_task INNER JOIN jobs_data ON
                    jobs_data.id = jobs_task.data_id
                WHERE
                    jobs_task.job_id = {self.id}
                    {'' if untagged else 'AND jobs_task.tag IS NOT NULL'}
                    {f"AND jobs_data.created_at >= '{start_date}'" if isinstance(start_date, str) else ''}
                    {f"AND jobs_data.created_at < '{end_date}'" if isinstance(end_date, str) else ''}
                """
            )
            n = cur.fetchone()[0]
        return n

    def get_by_data_id(self, id: int, cache=True, start_date=None, end_date=None):
        """
        Return task and tag using the data id
        """

        if id in self.cache:
            return self.cache[id]

        start_date = start_date or self.start_date
        end_date = end_date or self.end_date

        with self.db.conn.cursor() as cur:
            cur.execute(
                f"""
            SELECT
                jobs_data.data,
                jobs_task.tag,
                jobs_task.is_gold,
                jobs_task.tagged_time,
                jobs_data.id
            FROM jobs_task INNER JOIN jobs_data ON
                jobs_data.id = jobs_task.data_id
            WHERE
                jobs_task.job_id = {self.id} AND jobs_data.data_id = {id}
                {f"AND jobs_data.created_at >= '{start_date}'" if isinstance(start_date, str) else ''}
                {f"AND jobs_data.created_at < '{end_date}'" if isinstance(end_date, str) else ''}            
            """
            )
            try:
                task_dict, tag_list, is_gold, tagged_time = cur.fetchone()
            except TypeError:
                raise RuntimeError("No item found for given data id")

            task = build_task(task_dict, self.task_type, tz=self.tz)
            task.is_gold = bool(is_gold)
            tag = json.loads(tag_list)

            if cache:
                self.cache[id] = (task, tag, tagged_time)
            return task, tag, tagged_time

    def type(self):
        """
        List all available job_ids
        """
        import time
        time.sleep(.5)
        self.db = Database(
            db=self.db_name, user=self.user, password=self.password, host=self.host, port=self.port
        )
        with self.db.conn.cursor() as cursor:
            cursor.execute('select "taskType" from jobs_job where id = %s', (self.id,))
            record = cursor.fetchone()
            if record:
                return record[0]
            raise ValueError(f"Invalid job id {self.id}")

    def get_ids(
        self,
        untagged=False,
        itersize=1000,
        only_gold=False,
        start_date=None,
        end_date=None,
    ):
        """
        Return (generator) tagged tasks and tags from the database. Itersize sets
        the iteration size for the server sided cursor.

        If `untagged` is True, also return untagged items. This might be useful
        for checking, say, production metrics. If `only_gold` is True, return
        only items which are marked as gold.
        """
        start_date = start_date or self.start_date
        end_date = end_date or self.end_date

        query = f"""
        SELECT
            jobs_data.id
        FROM jobs_task INNER JOIN jobs_data ON
            jobs_data.id = jobs_task.data_id
        WHERE
            jobs_task.job_id = {self.id}
            {'' if untagged else 'AND jobs_task.tag IS NOT NULL'}
            {"AND jobs_task.is_gold = true" if only_gold else ''}
            {f"AND jobs_data.created_at >= '{start_date}'" if start_date else ''}
            {f"AND jobs_data.created_at < '{end_date}'" if end_date else ''}
        """
        with self.db.conn.cursor() as cur:
            cur.itersize = itersize
            cur.execute(query)
            data_ids = [x[0] for x in cur.fetchall()]
        self.db.conn.close()
        return data_ids


    def get(
        self,
        data_ids = [],
        untagged=False,
        itersize=1000,
        only_gold=False,
        start_date=None,
        end_date=None,
    ):
        """
        Return (generator) tagged tasks and tags from the database. Itersize sets
        the iteration size for the server sided cursor.

        If `untagged` is True, also return untagged items. This might be useful
        for checking, say, production metrics. If `only_gold` is True, return
        only items which are marked as gold.
        """
        start_date = start_date or self.start_date
        end_date = end_date or self.end_date

        query = f"""
        SELECT
            jobs_data.data,
            jobs_task.tag,
            jobs_task.is_gold,
            jobs_task.tagged_time,
            jobs_data.id
        FROM jobs_task INNER JOIN jobs_data ON
            jobs_data.id = jobs_task.data_id
        WHERE
            jobs_task.data_id IN {tuple(data_ids)}
        """
        db = Database(self.db_name, self.user, self.password, host=self.host, port=self.port)
        items = []
        with db.conn.cursor() as cur:
            cur.execute(query)

            for row in cur:
                task_dict, tag, is_gold, tagged_time, data_id = row
                task = build_task(task_dict, self.task_type, data_id, tz=self.tz)
                task.is_gold = bool(is_gold)
                items.append((task, tag, tagged_time))
        db.conn.close()
        return items


class LabelstudioJob(AbstractJob):
    """
    A Labelstudio job which specifies a kind of tagging data set and problem.
    """

    def __init__(
        self,
        id: int,
        task_type: str = "conversation",
        database: Optional[Database] = None,
        tz=pytz.UTC,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[Union[str, int]] = None,
    ):
        self.id = id
        self.db_name = db
        self.task_type = task_type
        self.db = database or Database(
            db=db, user=user, password=password, host=host, port=port
        )

        # Cache for keeping rows of job indexed by data_ids
        self.cache = {}
        self.tz = tz
        self.start_date = start_date
        self.end_date = end_date
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def total(self, untagged=False, start_date=None, end_date=None):
        """
        Return total number of items for this job. If `untagged` is True, consider
        untagged items in counting too.
        """
        start_date = start_date or self.start_date
        end_date = end_date or self.end_date

        with self.db.conn.cursor() as cur:
            cur.execute(
                f"""
                WITH labelstudio_data as (
                    SELECT 
                        project.id as "job_id",
                        task_completion.created_at as "task_completion_created_at"
                    FROM project
                    INNER JOIN task on project.id=task.project_id
                    INNER JOIN task_completion on task_completion.task_id=task.id
                    {"" if untagged else "WHERE task_completion.result != '[]'"}
                )
                
                SELECT COUNT(*) from labelstudio_data
                WHERE
                    job_id={self.id}
                    {f"AND task_completion_created_at >= '{start_date}'" if isinstance(start_date, str) else ''}
                    {f"AND task_completion_created_at <= '{end_date}'" if isinstance(end_date, str) else ''}
                """
            )
            n = cur.fetchone()[0]
        return n

    def type(self):
        """
        type of job
        """
        return ""
        

    def get_by_data_id(self, id: int, cache=True, start_date=None, end_date=None):
        """
        Return task and tag using the data id
        """

        if id in self.cache:
            return self.cache[id]

        start_date = start_date or self.start_date
        end_date = end_date or self.end_date

        with self.db.conn.cursor() as cur:
            cur.execute(
                f"""
                WITH labelstudio_data as (
                    SELECT 
                        project.id as "job_id",
                        task.data->> 'conversation_uuid' as "data_id",
                        task_completion.created_at as "task_completion_created_at",
                        task.data,
                        task_completion.result as "tag"
                    FROM project
                    INNER JOIN task on project.id=task.project_id
                    INNER JOIN task_completion on task_completion.task_id=task.id
                    {"" if untagged else "WHERE task_completion.result != '[]'"}
                )
                
                SELECT data_id, data, job_id, task_completion_created_at, tag FROM labelstudio_data
                WHERE
                    job_id={self.id} AND labelstudio_data.data_id={id}
                    {f"AND task_completion_created_at >= '{start_date}'" if isinstance(start_date, str) else ''}
                    {f"AND task_completion_created_at <= '{end_date}'" if isinstance(end_date, str) else ''}
                """
            )
            try:
                data_id, task_dict, job_id, tagged_time, tag_list = cur.fetchone()
            except TypeError:
                raise RuntimeError("No item found for given data id")

            task = build_task(task_dict, task_type='conversation', tz=self.tz)
            tag = json.loads(tag_list)[0]["value"]["choices"][0]

            if cache:
                self.cache[id] = (task, tag, tagged_time)
            return task, tag, tagged_time

    def get_ids(
        self,
        untagged=False,
        itersize=1000,
        only_gold=False,
        start_date=None,
        end_date=None,
    ):
        """
        Return (generator) tagged tasks and tags from the database. Itersize sets
        the iteration size for the server sided cursor.

        If `untagged` is True, also return untagged items. This might be useful
        for checking, say, production metrics. If `only_gold` is True, return
        only items which are marked as gold.
        """
        start_date = start_date or self.start_date
        end_date = end_date or self.end_date

        query = f"""
            WITH labelstudio_data as (
                SELECT 
                    project.id as "job_id",
                    task.data,
                    task_completion.created_at as "task_completion_created_at"
                FROM project
                INNER JOIN task on project.id=task.project_id
                INNER JOIN task_completion on task_completion.task_id=task.id
                {"" if untagged else "WHERE task_completion.result != '[]'"}
            )
            
            SELECT data->>'conversation_uuid' as "data_id" FROM labelstudio_data
            WHERE
                job_id={self.id}
                {f"AND task_completion_created_at >= '{start_date}'" if isinstance(start_date, str) else ''}
                {f"AND task_completion_created_at <= '{end_date}'" if isinstance(end_date, str) else ''}
            """
        with self.db.conn.cursor() as cur:
            cur.itersize = itersize
            cur.execute(query)
            data_ids = [x[0] for x in cur.fetchall()]
        self.db.conn.close()
        return data_ids


    def get(
        self,
        data_ids = [],
        untagged=False,
        itersize=1000,
        only_gold=False,
        start_date=None,
        end_date=None,
    ):
        """
        Return (generator) tagged tasks and tags from the database. Itersize sets
        the iteration size for the server sided cursor.

        If `untagged` is True, also return untagged items. This might be useful
        for checking, say, production metrics.
        """
        start_date = start_date or self.start_date
        end_date = end_date or self.end_date

        query = f"""
            WITH labelstudio_data as (
                SELECT 
                    project.id as "job_id",
                    task.data->> 'conversation_uuid' as "data_id",
                    task_completion.created_at as "task_completion_created_at",
                    task.data,
                    task_completion.result as "tag"
                FROM project
                INNER JOIN task on project.id=task.project_id
                INNER JOIN task_completion on task_completion.task_id=task.id
                {"" if untagged else "WHERE task_completion.result != '[]'"}
            )
            
            SELECT data_id, data, job_id, task_completion_created_at, tag FROM labelstudio_data
            WHERE
                job_id={self.id} AND data_id in {tuple(data_ids)}
                {f"AND task_completion_created_at >= '{start_date}'" if isinstance(start_date, str) else ''}
                {f"AND task_completion_created_at <= '{end_date}'" if isinstance(end_date, str) else ''}
            """
        
        db = Database(self.db_name, self.user, self.password, host=self.host, port=self.port)
        items = []
        with db.conn.cursor() as cur:
            cur.execute(query)

            for row in cur:
                data_id, task_dict, job_id, tagged_time, tag_list = row
                task = build_task(task_dict, "conversation", data_id, tz=self.tz)
                task.is_gold = True
                items.append((task, tag_list, tagged_time))
        db.conn.close()
        return items


class JobLocal(AbstractJob):
    """
    A tog job relying on local sqlite database.
    """

    def __init__(self, filepath: str, task_type="conversation", tz=pytz.UTC):
        self.task_type = task_type
        self.conn = sqlite3.connect(filepath)
        self.cache = {}
        self.tz = tz

    def total(self, untagged=False):
        """
        Return total number of items for this job. If `untagged` is True, consider
        untagged items in counting too.
        """

        cur = self.conn.cursor()
        cur.execute(
            f"SELECT count(*) FROM data {'' if untagged else 'WHERE tag IS NOT NULL'}"
        )
        return cur.fetchone()[0]

    def get_by_data_id(self, id: int, cache=True, show_source=False):
        """
        Return task and tag using the data id
        NOTE: We are not using cache here.
        """

        on_source = ", source" if show_source else ""

        cur = self.conn.cursor()
        cur.execute(
            f"""SELECT
          data, tag, is_gold, tagged_time{on_source}
        FROM data
        WHERE data_id = '{id}'
        """
        )

        try:
            if show_source:
                task_dict, tag_list, is_gold, tagged_time, source = cur.fetchone()
            else:
                task_dict, tag_list, is_gold, tagged_time = cur.fetchone()
        except TypeError:
            raise RuntimeError("No item found for given data id")

        task = build_task(json.loads(task_dict), self.task_type, tz=self.tz)
        task.is_gold = bool(is_gold)
        tag = json.loads(tag_list)
        if show_source:
            return task, tag, tagged_time, source
        return task, tag, tagged_time

    def get(
        self,
        untagged=False,
        itersize=1000,
        only_gold=False,
        show_source=False,
        show_ids=False,
    ):
        """
        Return (generator) tagged tasks and tags from the database.

        If `untagged` is True, also return untagged items. This might be useful
        for checking, say, production metrics. If `only_gold` is True, return
        only items which are marked as gold.
        """
        cur = self.conn.cursor()
        on_source = ", source" if show_source else ""
        on_ids = "data_id," if show_ids else ""
        cur.execute(
            f"""SELECT
          {on_ids} data, tag, is_gold, tagged_time {on_source}
        FROM data
        {'' if untagged else 'WHERE tag IS NOT NULL'}
        {'AND is_gold = 1' if only_gold else ''}
        """
        )

        for row in cur:
            if show_source and show_ids:
                id_, task_dict, tag, is_gold, tagged_time, source = row
            elif show_source:
                task_dict, tag, is_gold, tagged_time, source = row
            elif show_ids:
                id_, task_dict, tag, is_gold, tagged_time = row
            else:
                task_dict, tag, is_gold, tagged_time = row
            task = build_task(json.loads(task_dict), self.task_type, tz=self.tz)
            task.is_gold = bool(is_gold)
            if show_source and show_ids:
                yield id_, task, json.loads(tag), tagged_time, source
            elif show_source:
                yield task, json.loads(tag), tagged_time, source
            elif show_ids:
                yield id_, task, json.loads(tag), tagged_time
            else:
                yield task, json.loads(tag), tagged_time



import ast
import uuid
import asyncio
import ast
import json
import os
import tempfile
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union
import aiofiles

import aiohttp
import attr
import dvc.api
import jsonschema
import numpy as np
import pandas as pd
import pytz
from loguru import logger
from requests import JSONDecodeError
from tqdm import tqdm

from skit_labels import constants as const
from skit_labels.db import Database, Job, LabelstudioJob, SqliteDatabase


def batch_gen(source, n=100):
    """
    Batched generator
    """

    batch = []

    for it in source:
        batch.append(it)
        if len(batch) % n == 0:
            yield batch
            batch = []

    if batch:
        yield batch


def download_dataset(
    job_id: str,
    task_type: str,
    timezone: pytz.BaseTzInfo = pytz.UTC,
    full: bool = False,
    batch_size: int = 500,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
) -> Tuple[SqliteDatabase, str, str]:
    database = Database(db=db, user=user, password=password, host=host, port=port)
    JOB_CREATOR = LabelstudioJob if db == const.LABELSTUIO_DB else Job
    job = JOB_CREATOR(
        int(job_id),
        task_type=task_type,
        tz=timezone,
        start_date=start_date,
        end_date=end_date,
        database=database,
        db=db,
        user=user,
        password=password,
        host=host,
        port=port
    )
    if db != const.LABELSTUIO_DB:
        describe_dataset(job_id, job=job)
        stat_dataset(job_id, job=job)

    _, temp_filepath = tempfile.mkstemp(suffix=const.OUTPUT_FORMAT__SQLITE)
    sdb = SqliteDatabase(temp_filepath)
    bar = tqdm(total=job.total(untagged=full))
    data_ids = job.get_ids(untagged=full, start_date=start_date, end_date=end_date)
    
    for start_index in range(0, len(data_ids), batch_size):
        items = job.get(data_ids=data_ids[start_index:start_index+batch_size], untagged=full, start_date=start_date, end_date=end_date)
        rows = []
        for task, tag, tagged_time in items:
            # For raw dictionary type tasks, we don't use attr classes.
            task_dict = task if isinstance(task, dict) else attr.asdict(task)

            # TODO: is_gold might not be working for dict type tasks as of now
            rows.append(
                (
                    task.id,
                    task_dict,
                    tag,
                    task.is_gold,
                    tagged_time,
                    job_id
                )
            )

        sdb.insert_rows(rows)
        bar.update(n=len(items))
    return sdb, temp_filepath, job.type()


def parse_json(data: str) -> Dict[str, Any]:
    try:
        data = json.loads(data)
        data = data if isinstance(data, dict) else json.loads(data)
        return data
    except JSONDecodeError:
        return {}


def unpack(df: pd.DataFrame) -> pd.DataFrame:
    df.data = df.data.apply(parse_json)
    df_dict = df.to_dict(orient="records")
    df = pd.json_normalize(df_dict)
    columns = {col: col.replace("data.", "") for col in df.columns if col.startswith("data.") and "data_id" not in col}
    df.rename(columns=columns, inplace=True)
    return df


def sdb2df(sdb: SqliteDatabase, job_id: str) -> str:
    _, output_file = tempfile.mkstemp(
        prefix=f"job-{job_id}-", suffix=const.OUTPUT_FORMAT__CSV
    )
    df = pd.read_sql_query("SELECT * FROM data", sdb.conn)
    df = unpack(df)
    df.to_csv(output_file, index=False)
    return output_file


def describe_dataset(
    job_id: Optional[int] = None,
    job: Optional[Job] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
) -> str:
    return job or Job(
        int(job_id),
        db=db,
        user=user,
        password=password,
        host=host,
        port=port,
    )


def stat_dataset(
    job_id: Optional[int] = None,
    job: Optional[Job] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
) -> dict:
    job_ = job or Job(
        int(job_id),
        start_date=start_date,
        end_date=end_date,
        db=db,
        user=user,
        password=password,
        host=host,
        port=port,
    )
    n_total = job_.total(untagged=True)
    n_tagged = job_.total()
    return {
        const.TOTAL: n_total,
        const.TAGGED: n_tagged,
        const.UNTAGGED: n_total - n_tagged,
    }


def print_job_stats(job_stats: dict) -> str:
    return json.dumps(job_stats, indent=2)


def download_dataset_from_dvc(
    repo: str, path: str, remote: Optional[str] = None
) -> str:
    file_name = os.path.split(path)[-1]
    _, output_file = tempfile.mkstemp(suffix=file_name)
    with dvc.api.open(path, repo=repo, remote=remote) as f:
        df = pd.read_csv(f)
        df.to_csv(output_file, index=False)
    return output_file


def processLabelstudioColumns(df_path: str):
    df = pd.read_csv(df_path)
    df[const.DATA_ID] = df[const.CONVERSATION_UUID].values
    df[const.ALTERNATIVES] = df[const.UTTERANCES].values if const.UTTERANCES in df else df[const.ALTERNATIVES]
    df[const.ALTERNATIVES] = df[const.ALTERNATIVES].apply(
        lambda val: json.dumps(json.loads(json.loads(val)), ensure_ascii=False) \
            if json.loads(val) else json.dumps([])
    )
    try:
        df["tag"] = df["tag"].apply(lambda val: json.dumps(json.loads(val)[0]["value"]))
        df = df[df["tag"].apply(lambda val: "choices" in json.loads(val))]
        df["tag"] = df["tag"].apply(lambda val: json.loads(val)["choices"][0])
    except json.JSONDecodeError:
        logger.warning("please check tag column, it's unparseable to get a single value out")
        
    df.to_csv(df_path, index=False)

async def download_dataset_from_labelstudio(
    url: str,
    token: str,
    project_id: Union[int, str]
) -> Tuple[str, str]:
    """
    Download dataset from labelstudio
    """
    _, output_file = tempfile.mkstemp(suffix=const.OUTPUT_FORMAT__CSV)
    headers = {
        "Authorization": f"token {token}",
    }
    async with aiohttp.ClientSession(url, headers=headers) as session:
        async with session.get(url=f"/api/projects/{project_id}/export?exportType=CSV") as response:
            if response.status != 200:
                error_message = await response.text()
                raise Exception(f"Error downloading dataset: {error_message} {response.status} ")
            async with aiofiles.open(output_file, mode='wb') as f:
                await f.write(await response.read())
    
    processLabelstudioColumns(df_path=output_file)
    return output_file, "csv"


def download_dataset_from_db(
    job_id: str,
    task_type: str,
    timezone: pytz.BaseTzInfo = pytz.UTC,
    full: bool = False,
    batch_size: int = 500,
    output_format: str = const.OUTPUT_FORMAT__CSV,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
) -> Tuple[str, str]:
    sdb, sdb_path, dataset_type = download_dataset(
        job_id,
        task_type,
        timezone,
        full,
        batch_size,
        start_date=start_date,
        end_date=end_date,
        db=db,
        user=user,
        password=password,
        host=host,
        port=port,
    )
    if output_format == const.OUTPUT_FORMAT__CSV:
        df_path = sdb2df(sdb, job_id)
        if db == const.LABELSTUIO_DB:
            processLabelstudioColumns(df_path)
        os.remove(sdb_path)
        return df_path, dataset_type
    else:
        return sdb_path, dataset_type


def extract_utterances_safely(conversation_uuid, utterances):
    if not isinstance(utterances, (list, str)):
        return []
    try:
        utterances = (
            json.loads(utterances) if isinstance(utterances, str) else utterances
        )
    except json.JSONDecodeError:
        utterances = ast.literal_eval(utterances) if isinstance(utterances, str) else []
    except Exception as e:
        logger.warning(
            f"{conversation_uuid} has invalid utterances: {utterances}, setting to []."
        )
        logger.warning(e)
        utterances = []
    return utterances


def build_dataset(
    job_id: str, data_frame: pd.DataFrame, source: Optional[str] = const.DEFAULT_SOURCE
) -> List[dict]:
    """
    Build a dataset from the dataframe.

    :param job_id: The dataset id where data should be uploaded.
    :type job_id: str
    :param data_frame: The dataframe to upload.
    :type data_frame: pd.DataFrame
    :param source: The source of the data_frame.
    :type source: Optional[str]
    :return: The job-id where the data was uploaded.
    :rtype: int
    """
    dataset = []
    logger.debug(f"Pushing {len(data_frame)} items to {job_id=}")
    data_frame.fillna(np.nan, inplace=True)
    data_frame.replace([np.nan], [None], inplace=True)
    for _, row in tqdm(
        data_frame.iterrows(),
        total=len(data_frame),
        desc="Building a dataset for uploading safely.",
    ):

        conversation_uuid = row[const.CONVERSATION_UUID]
        dedupe_id = f"{conversation_uuid}_{uuid.uuid4().hex}"
        errors = []
        if const.RAW in data_frame.columns:
            data = json.loads(row[const.RAW])
        else:
            data = row.to_dict()
        utterance_columns = {const.UTTERANCES, const.ALTERNATIVES}

        if data_frame.columns.intersection(utterance_columns).empty:
            raise ValueError(f"Expected one of {const.UTTERANCES} or {const.ALTERNATIVES} "
            "columns in the dataframe. {data_frame.columns}")
        utterance_col = const.UTTERANCES if const.UTTERANCES in data_frame.columns else const.ALTERNATIVES

        data_point = {
            const.PRIORITY: 1,
            const.DATA_SOURCE: source,
            const.DATA_ID: dedupe_id,
            const.DATA: {
                **data,
                const.CALL_UUID: str(row[const.CALL_UUID]),
                const.CONVERSATION_UUID: str(row[const.CONVERSATION_UUID]),
                const.ALTERNATIVES: extract_utterances_safely(row[const.CONVERSATION_UUID], row[utterance_col]),
            },
            const.IS_GOLD: False,
        }
        try:
            jsonschema.validate(data_point[const.DATA], const.UPLOAD_DATASET_SCHEMA)
            dataset.append(data_point)
        except jsonschema.exceptions.ValidationError as e:
            errors.append(e)
            if len(errors) > len(data_frame) * 0.5:
                raise RuntimeError(f"Too many errors: {errors}")

    return dataset


async def upload_dataset(
    session: aiohttp.ClientSession, job_id: str, dataset: List[dict], retries: int = 3
):
    sleep_time = 5 #seconds
    while retries >= 0 :
        path = f"/tog/tasks/?job_id={job_id}"
        status_code = 0
        try:
            async with session.post(path, json=dataset) as response:
                status_code = response.status
                if str(response.status).startswith("2"):
                    upload_response = await response.json()
                    return (upload_response, status_code)
                else:
                    raise aiohttp.ClientOSError
                
        except (aiohttp.ClientOSError, aiohttp.ServerDisconnectedError, asyncio.TimeoutError) as e:
            retries -= 1
            print(f"failed to upload dataset: {e},\n..retrying in {sleep_time} seconds")
            await asyncio.sleep(sleep_time)
    
    raise Exception(f"Error uploading dataset: {status_code}")


async def upload_dataset_batches(
    dataset_chunks: Iterable[List[dict]], url: str, token: str, job_id: str
) -> str:
    """
    Post the dataset to the server.

    :param dataset_chunks: The dataset to post.
    :type dataset_chunks: Iterable[dict]
    :param token: The token to use for authentication.
    :type token: str
    :param job_id: The job id where the dataset should be uploaded.
    :type job_id: str
    :return: The job id where the dataset was uploaded.
    :rtype: int
    """
    headers = {"Authorization": f"Bearer {token}"}
    print("Uploading batches")
    async with aiohttp.ClientSession(url, headers=headers) as session:
        requests = [
            upload_dataset(session, job_id, dataset) for dataset in dataset_chunks
        ]
        return await asyncio.gather(*requests)


async def upload_file(input_file: str, project_id: str, session: aiohttp.ClientSession) -> str:
    with open(input_file, "rb") as f:
        return await session.post(f"/api/projects/{project_id}/import", data={"file": f})


async def upload_dataset_to_labelstudio(
    input_file: str,
    url: str,
    token: str,
    project_id: str
) -> Tuple[List[str], int]:
    """
    Upload the dataset to LabelStudio.

    :return: The job id where the dataset was uploaded.
    :rtype: int
    """
    headers = {"Authorization": f"token {token}"}
    async with aiohttp.ClientSession(url, headers=headers) as session:
        response = await upload_file(input_file, project_id, session)

        if response.status != 201:
            error_message = await response.text()
            raise RuntimeError(f"Failed to upload dataset to LabelStudio: {error_message}, {response.status}")
        else:
            response = await response.json()
            return [], response["task_count"]


async def upload_dataset_to_db(
    input_file: str, url: str, token: str, job_id: str
) -> Tuple[List[str], int]:
    """
    Uploads a dataset to the database.

    :param input_file: Path to the input file.
    :type input_file: str
    :param url: The url to the dataset server.
    :type url: str
    :param token: The token for uploading to the target organization.
    :type token: str
    :param job_id: The dataset id where data should be uploaded, defaults to None
    :type job_id: Optional[str], optional
    :return: The job-id where the data was uploaded.
    :rtype: str
    """
    _, extension = os.path.splitext(input_file)

    if extension != ".csv":
        raise ValueError("Expected file extension to be a csv.")

    data_frame = pd.read_csv(input_file)
    dataset = build_dataset(job_id, data_frame)
    batched_datasets = batch_gen(dataset, 100)
    errors_final = []
    for batched_dataset in batch_gen(batched_datasets, 10):
        responses = await upload_dataset_batches(batched_dataset, url, token, job_id)
        errors = []

        for message, status_code in responses:
            if status_code not in [200, 201]:
                errors.append(message)
                logger.error(f"{status_code}: {message}")
        errors_final.extend(errors)
    return errors_final, len(data_frame)



"""
Module provides access to logger config, session token and package version.
"""
import os
import sys

import toml
from typing import Optional
from loguru import logger
from datetime import datetime
import pandas as pd
from typing import Union


LOG_LEVELS = ["CRITICAL", "ERROR", "WARNING", "SUCCESS", "INFO", "DEBUG", "TRACE"]


def get_version():
    project_toml = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "pyproject.toml")
    )
    with open(project_toml, "r") as handle:
        project_metadata = toml.load(handle)
    return project_metadata["tool"]["poetry"]["version"]


def configure_logger(level: int) -> None:
    """
    Configure the logger.
    """
    size = len(LOG_LEVELS)
    if level >= size:
        level = size - 1
    log_level = LOG_LEVELS[level]

    config = {
        "handlers": [
            {
                "sink": sys.stdout,
                "format": """
    -------------------------------------------------------
    <level>{level}</level>
    -------
    TIME: <green>{time}</green>
    FILE: {name}:L{line} <blue>{function}(...)</blue>
    <level>{message}</level>
    -------------------------------------------------------
    """,
                "colorize": True,
                "level": log_level,
            },
            {
                "sink": "file.log",
                "rotation": "500MB",
                "retention": "10 days",
                "format": "{time} {level} -\n{message}\n--------------------\n",
                "level": log_level,
            },
        ],
    }
    logger.configure(**config)
    logger.enable(__name__)


def read_session() -> Optional[str]:
    """
    Read the session from the environment.
    """
    home = os.path.expanduser("~")
    try:
        with open(os.path.join(home, ".skit", "token"), "r") as handle:
            return handle.read().strip()
    except FileNotFoundError:
        return None


def to_datetime(d: str) -> datetime:
    """
    Convert an arbitrary date string to ISO format.

    :param d: A date string.
    :type d: str
    :raises ValueError: If the date string can't be parsed.
    :return: The date string in ISO format.
    :rtype: datetime
    """
    if not isinstance(d, str):
        raise TypeError(f"Expected a string, got {type(d)}")

    time_fns = [
        datetime.fromisoformat,
        lambda date_string: datetime.strptime(
            date_string, "%Y-%m-%d %H:%M:%S.%f %z %Z"
        ),
        lambda date_string: datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ"),
        lambda date_string: datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%f%z"),
    ]

    for time_fn in time_fns:
        try:
            return time_fn(d)
        except ValueError:
            continue
    return None


def add_data_label(input_file: str, data_label: Optional[str] = None) -> str:
    df = pd.read_csv(input_file)
    data_label = data_label or None
    df = df.assign(data_label=data_label)
    df.to_csv(input_file, index=False)
    return input_file
